use nqp;
use Dawa::Debugger;
use QAST:from<NQP>;
use Terminal::ANSI::OO 't';

my Bool %debugging;
my %tracking;

my %breakpoints;

my $debugger = Dawa::Debugger.new;

class TrackingState {
  has $.context;
  has $!backtrace;
  has $.thread-gist = $*THREAD.gist;
  has $.file;
  has $.line;

  method backtrace is hidden-from-backtrace { $!backtrace }
  method TWEAK is hidden-from-backtrace {
    $!backtrace //= Backtrace.new;
    with $!backtrace.list.first( { !.is-setting && !.is-hidden } ) {
      $!file = .file;
      $!line = .line;
    }
  }
  method Str {
    join "\n",
     $.thread-gist, |$.backtrace.list.grep( {
       !.is-setting && !.is-hidden
     } ).map: {
       .Str.trim-trailing
     }
  }
}

sub stop is export is hidden-from-backtrace {
  $debugger.stop-thread;
  %debugging{ $*THREAD.id } = True;
  %tracking{ $*THREAD.id } = TrackingState.new;
}

sub track is export is hidden-from-backtrace {
  %tracking{ $*THREAD.id } = TrackingState.new;
}

my Lock $repl-lock .= new;

sub maybe-stop($context) is hidden-from-backtrace {
  stop if $debugger.breakpoint(callframe(1).file,callframe(1).line);
  return unless %debugging{ $*THREAD.id };
  my $stack = Backtrace.new;
  $repl-lock.protect: {
    $debugger.run-repl(:$context,:$stack, :%tracking);
  }
  $debugger.update-state(:%debugging);
  %tracking{ $*THREAD.id } = TrackingState.new;
};

sub EXPORT(|) {
  role Dawa {
    method statement(Mu $/) {
      $/.make:
        QAST::Stmts.new(
          QAST::Op.new( :op('call'), QAST::WVal.new( :value(&maybe-stop) ),
            # pseudostash:
            QAST::Op.new(
               :op('callmethod'), :name('new'),
               QAST::WVal.new( :value($*W.find_single_symbol('PseudoStash')))
            )
          ),
          callsame
        );
    }
  }
  $*LANG.define_slang: 'MAIN', $*LANG.slang_grammar('MAIN'), $*LANG.actions.^mixin(Dawa);
  {}
}
